/*
    ext2prepare.c -- prepares an ext2 fs for an online resize
    Copyright (C) 1999 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This program is to be used in cooperation with Andreas Dilger's
  online resize patch for the linux kernel.
*/

static const char _ext2prepare_c[] = "$Id: ext2prepare.c,v 1.4 1999/08/22 20:35:19 buytenh Exp $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include "ext2.h"

char *prog;

void usage(FILE *outfile)
{
	fprintf(outfile, "usage: %s device max_new_size[kmgt]\n", prog);
	fprintf(outfile,
"       where max_new_size is in blocks, Kilo-, Mega-, Giga-, or Terabytes\n");
}

blk_t getfreezeroedblock(struct ext2_fs *fs)
{
	int i;
	blk_t offset;

	for (i = 0, offset=fs->sb.s_first_data_block; i < fs->numgroups;
	     i++, offset += fs->sb.s_blocks_per_group)
	{
		blk_t j;

		for (j = fs->itoffset + fs->inodeblocks;
		     j < fs->sb.s_blocks_per_group; j++)
			if (!ext2_get_block_state(fs, offset + j))
			{
				struct ext2_buffer_head *bh;

				ext2_set_block_state(fs, offset + j, 1, 1);
				bh = ext2_bcreate(fs, offset + j);
				bh->dirty = 1;
				ext2_brelse(bh, 0);

				return offset + j;
			}
	}

	return 0;
}

int putblk(struct ext2_fs *fs, struct ext2_inode *inode, blk_t block)
{
	struct ext2_buffer_head *bh;
	__u32			*udata;
	int			 i;
	int			 u32perblock = fs->blocksize >> 2;

	if (fs->opt_verbose)
		fprintf(stderr, "adding block %u to inode\n", block);

	/* Direct blocks for first 12 blocks */
	for (i = 0; i < EXT2_NDIR_BLOCKS; i++)
		if (inode->i_block[i] == block)
			return 1;
		else if (!inode->i_block[i])
		{
			inode->i_block[i] = block;
			inode->i_blocks += 2;	/* Why 2?  Ask e2fsck! */
			inode->i_size += fs->blocksize;
			ext2_set_block_state(fs, block, 1, 1);
			return 1;
		}

	if (!inode->i_block[EXT2_IND_BLOCK])
	{
		if (!(inode->i_block[EXT2_IND_BLOCK] = getfreezeroedblock(fs)))
			return 0;

		inode->i_blocks += 2;
	}
	bh = ext2_bread(fs, inode->i_block[EXT2_IND_BLOCK]);
	udata = (__u32 *)bh->data;

	/* Indirect blocks for next 256 1k blocks */
	for (i = 0; i < u32perblock; i++)
		if (udata[i] == block)
		{
			ext2_brelse(bh, 0);
			return 1;
		}
		else if (!udata[i])
		{
			bh->dirty = 1;
			udata[i] = block;
			inode->i_blocks += 2;
			inode->i_size += fs->blocksize;
			ext2_set_block_state(fs, block, 1, 1);
			ext2_brelse(bh, 0);
			return 1;
		}

	ext2_brelse(bh, 0);

	if (!inode->i_block[EXT2_DIND_BLOCK])
	{
		if (!(inode->i_block[EXT2_DIND_BLOCK] = getfreezeroedblock(fs)))
			return 0;

		inode->i_blocks += 2;
	}
	bh = ext2_bread(fs, inode->i_block[EXT2_DIND_BLOCK]);
	udata = (__u32 *)bh->data;

	/* Double indirect blocks for next 65536 1k blocks */
	for (i = 0; i < u32perblock; i++)
	{
		struct ext2_buffer_head	*bh2;
		__u32			*udata2;
		int			 j;

		if (!udata[i])
		{
			if (!(udata[i] = getfreezeroedblock(fs)))
			{
				ext2_brelse(bh, 0);
				ext2_brelse(bh2, 0);
				return 0;
			}
			bh->dirty = 1;
			inode->i_blocks += 2;
		}
		bh2 = ext2_bread(fs, udata[i]);
		udata2 = (__u32 *)bh2->data;

		for (j = 0; j < u32perblock; j++)
			if (udata2[j] == block)
			{
				ext2_brelse(bh, 0);
				ext2_brelse(bh2, 0);
				return 1;
			}
			else if (!udata2[j])
			{
				bh2->dirty = 1;
				udata2[j] = block;
				ext2_set_block_state(fs, block, 1, 1);
				inode->i_blocks += 2;
				inode->i_size += fs->blocksize;
				ext2_brelse(bh, 0);
				ext2_brelse(bh2, 0);
				return 1;
			}
		ext2_brelse(bh2, 0);
	}
	ext2_brelse(bh, 0);

	return 0;
}

int create_resize_inode(struct ext2_fs *fs, blk_t newsize)
{
	int			diff;
	int			i;
	struct ext2_inode	inode;
	int			newgroups;
	int			newgdblocks;
	blk_t			block;

	newgroups = howmany(newsize - fs->sb.s_first_data_block, fs->sb.s_blocks_per_group);
	newgdblocks = howmany(newgroups * sizeof(struct ext2_group_desc), fs->blocksize);

	ext2_read_inode(fs, EXT2_RESIZE_INO, &inode);
	if (inode.i_mode == 0)
	{
		if (fs->opt_verbose)
			fprintf(stderr,
				"creating inode %03d\n",
				EXT2_RESIZE_INO);

		inode.i_mode = S_IFREG | S_IRUSR | S_IWUSR;
		inode.i_size = 0;
		inode.i_uid = inode.i_gid = 0;
		inode.i_atime = inode.i_ctime = time(NULL);
		inode.i_dtime = 0;
		inode.i_links_count = 1;
		inode.i_blocks = 0;
		inode.i_flags = EXT2_NODUMP_FL;
		inode.osd1.linux1.l_i_reserved1 = 0;
		for (i = 0; i < EXT2_NDIR_BLOCKS; i++)
			inode.i_block[i] = 0;
		inode.i_block[EXT2_IND_BLOCK] = 0;
		inode.i_block[EXT2_DIND_BLOCK] = 0;
		inode.i_block[EXT2_TIND_BLOCK] = 0;
		inode.i_generation = 0;
		inode.i_file_acl = 0;
		inode.i_dir_acl = 0;
		inode.i_faddr = 0;
		inode.osd2.linux2.l_i_frag = 0;
		inode.osd2.linux2.l_i_fsize = 0;
		inode.osd2.linux2.i_pad1 = 0;
		inode.osd2.linux2.l_i_reserved2[0] = 0;
		inode.osd2.linux2.l_i_reserved2[1] = 0;
	}

	diff = newgdblocks - fs->gdblocks;

	for (i = 0, block = fs->sb.s_first_data_block + fs->gdblocks + 1;
	     i < fs->numgroups; i++, block += fs->sb.s_blocks_per_group)
	{
		int j;
		for (j = 0; j < diff; j++)
			if (!putblk(fs, &inode, block + j))
				return 0;
	}

	inode.i_mtime = time(NULL);
	ext2_write_inode(fs, EXT2_RESIZE_INO, &inode);

	return 1;
}


int ext2_prepare_fs(struct ext2_fs *fs, blk_t maxnewsize)
{
	int status;

	if (fs->opt_verbose)
		fprintf(stderr, "%s: resize to %u blocks\n", prog, maxnewsize);

	if (maxnewsize <= fs->sb.s_blocks_count)
		return 1;

	if ((fs->relocator_pool = (unsigned char *)malloc(ext2_relocator_pool_size << 10)) == NULL)
	{
		fprintf(stderr, "%s: error allocating relocator pool!\n", prog);
		return 0;
	}
	fs->relocator_pool_end = fs->relocator_pool + (ext2_relocator_pool_size << 10);

	status = ext2_block_relocate(fs, maxnewsize);

	if (status)
		status = ext2_metadata_push(fs, maxnewsize);

	if (status)
		status = create_resize_inode(fs, maxnewsize);

	free(fs->relocator_pool);
	fs->relocator_pool = NULL;
	fs->relocator_pool_end = NULL;

	return status;
}

int main(int argc, char *argv[])
{
	struct ext2_dev_handle *handle;
	struct ext2_fs	       *fs;
	int		        newgroups;
	blk_t		        newsize;
	char			mod = '\0';

	prog = argv[0];

	if (argc != 3)
	{
		usage(stderr);
		return 1;
	}

	handle = ext2_make_dev_handle_from_file(argv[1]);
	if (handle == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", argv[1], prog);
		return 1;
	}

	sscanf(argv[2], "%i%c", &newsize, &mod);
	switch (mod)
	{
		case 't':
		case 'T': newsize *= 1024;
		case 'g':
		case 'G': newsize *= 1024;
		case 'm':
		case 'M': newsize *= 1024;
		case 'k':
		case 'K': /* Need to load fs to find block size */
		case  0 :
		case 'b':
		case 'B': /* do nothing */ break;
		default: usage(stderr);
	}

	newgroups = howmany(newsize, 1024); /* 1k block size is worst case */
	if (ext2_max_groups < newgroups)
		ext2_max_groups = newgroups;

	if ((fs = ext2_open(handle, 0)) == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", argv[1], prog);
		return 1;
	}

	if (mod && mod != 'b' && mod != 'B')
	{
		newsize /= fs->blocksize / 1024;
	}

	if (newsize >= 1<<28 && fs->blocksize == 1024)
	{
		fprintf(stderr, "%s: can't resize 1k block filesystem over 256GB.\n", prog);
		fprintf(stderr, "%s: too much space would be wasted, sorry.\n", prog);
		return 1;
	}

	if (!ext2_prepare_fs(fs, newsize))
		return 1;

	ext2_close(fs);

	return 0;
}

